﻿;/////////////////////////////////////////////////////////////////////////////////
;***COMate***  COM automation through iDispatch.
;*===========
;*
;*Excel demo.
;/////////////////////////////////////////////////////////////////////////////////

IncludePath "..\"
XIncludeFile "COMatePLUS.pbi"

;Set the following constant to a string identifier which will be used to identify the instance of Excel.
;This must be unique amongst all possible instances.
#ExcelCaption$ = "COMate_EmbedExcel"

EnableExplicit

Global gOldProc, excelObject.COMateObject

Define eventID, excelhWnd


Enumeration
  #ExcelContainer
EndEnumeration


;/////////////////////////////////////////////////////////////////////////////////
;The following is called repeatedly in order to locate the Excel window.
Procedure.i enumWindowProc(hWnd, *excelhWnd.INTEGER)
  Protected title$, hMenu, i
  title$ = Space(100)
  GetWindowText_(hWnd, @title$, 100)
  If title$ = #ExcelCaption$ And IsWindowVisible_(hWnd) = 0
    *excelhWnd\i = hWnd
    ProcedureReturn 0
  EndIf
  ProcedureReturn 1
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i LoadExcel(hWndContainer)
  Protected excelhWnd, hMenu, i, rc.RECT
  Protected borderWidth, borderHeight, captionHeight, statusHeight
  excelObject = COMate_CreateObject("Excel.Application")
  If excelObject
    excelObject\SetProperty("Caption = '" + #ExcelCaption$ + "'")
    EnumWindows_(@enumWindowProc(), @excelhWnd)
    If excelhWnd
      excelObject\SetProperty("Visible=#True")
      SetGadgetData(#ExcelContainer, excelhWnd)
      ;Disable all system menu items.
        hMenu = GetSystemMenu_(excelhWnd, #False)
        For i = GetMenuItemCount_(hMenu)-1 To 0 Step -1
          RemoveMenu_(hMenu, i, #MF_REMOVE|#MF_BYPOSITION) 
        Next
      ;Place the window inside the container and reposition / resize it.
        SetParent_(excelhWnd, GadgetID(0))
        SendMessage_(hWndContainer, #WM_SIZE, 0, 0)


      ProcedureReturn excelObject
    Else
      excelObject\Release()
    EndIf
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;The callback is needed to repaint the Excel window to combat problems; a bug with the PB webgadget?
Procedure.i containerCallback(hWnd, uMsg, wParam, lParam)
  Protected result, excelhWnd, rc.RECT, win
  Protected borderWidth, borderHeight, captionHeight, statusHeight
  Select uMsg
    Case #WM_LBUTTONDOWN
      Debug 4
    Case #WM_SIZE
      excelhWnd = GetGadgetData(#ExcelContainer)
      GetClientRect_(GadgetID(#ExcelContainer), rc)
      borderWidth = GetSystemMetrics_(#SM_CXFRAME)
      borderHeight = GetSystemMetrics_(#SM_CYFRAME)
      captionHeight = GetSystemMetrics_(#SM_CYCAPTION)
      statusHeight = GetSystemMetrics_(#SM_CYSMCAPTION)
      MoveWindow_(excelhWnd, -borderWidth, -borderHeight-captionHeight, rc\right-rc\left+2*borderWidth, rc\bottom-rc\top+2*borderHeight+captionHeight+statusHeight, 1)
    Case #WM_PAINT
      result = CallWindowProc_(gOldProc, hWnd, uMsg, wParam, lParam)
      excelhWnd = GetGadgetData(#ExcelContainer)
      If excelhWnd
        RedrawWindow_(excelhWnd, 0, 0, #RDW_INVALIDATE|#RDW_FRAME|#RDW_ALLCHILDREN)
      EndIf
    Case #WM_ERASEBKGND
      result = 1
    Default
      result = CallWindowProc_(gOldProc, hWnd, uMsg, wParam, lParam)
  EndSelect
  ProcedureReturn Result
EndProcedure


If OpenWindow(0, 0, 0, 800, 600, "Embed Excel application", #PB_Window_SystemMenu | #PB_Window_ScreenCentered | #PB_Window_SizeGadget | #PB_Window_MaximizeGadget | #PB_Window_MinimizeGadget)
  ;Create a container to house the Excel window. Ideally this container should be subclassed and all erasing prevented etc.
    ContainerGadget(#ExcelContainer, 250, 0, 0, 0, #PB_Container_BorderLess)
    gOldProc = SetWindowLong_(GadgetID(#ExcelContainer), #GWL_WNDPROC, @containerCallback())

  ;Load Excel into our container.
    excelObject = LoadExcel(GadgetID(#ExcelContainer))
    If excelObject = 0
      MessageRequester("COMate - Embed Excel application demo", "There was a problem loading Excel!")
      End
    EndIf

  Repeat
    eventID = WaitWindowEvent()
    Select eventID
      Case #WM_SIZE
        ResizeGadget(#ExcelContainer, #PB_Ignore, 0, WindowWidth(0)-250, WindowHeight(0))
    
    EndSelect
  Until eventID = #PB_Event_CloseWindow
  HideWindow(0, 1)  ;Avoids flicker when releasing the Excel objects.

  ;Avoid crashes.
    excelhWnd = GetGadgetData(#ExcelContainer)
    ShowWindow_(excelhWnd, #SW_HIDE)
    SetParent_(excelhWnd, 0)
  ;Tidy up.
    excelObject\Invoke("Quit()")
    excelObject\Release()
EndIf

End

; IDE Options = PureBasic 4.30 (Windows - x86)
; ExecutableFormat = Shared Dll
; CursorPosition = 8
; FirstLine = 2
; Folding = -
; EnableUnicode
; EnableThread
; EnableXP
; Executable = nxReportU.dll